/*
 * KANTS - KDE Applet for the ANTS system
 *
 * (C) 2001 by Klaus S. Madsen  (ksm@sysorb.com / ksm@hjernemadsen.org)
 *
 * $Id: kantsapplet.cpp,v 1.2 2003/03/17 19:21:20 jcd Exp $
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later
 *  version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA
 *
 */
#include <klocale.h>
#include <kglobal.h>
#include <kaboutapplication.h>
#include <qpopupmenu.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <algorithm>
#include <qpainter.h>
#include "kantsapplet.h"
#include "kantsapplet.moc"
#include "../src/config.h"

const int TEXT_MARGIN = 6;
const int TEXT_SIZE = 8;
const int BAR_MARGIN = 2;
const int LOW_CELLSIZE = 3;

extern "C"
{
  KAntsApplet* init (QWidget *parent, const QString& configFile)
  {
    KGlobal::locale()->insertCatalogue("kantsapplet");
    return new KAntsApplet (configFile, KPanelApplet::Normal,
			    KPanelApplet::About, parent, "kantsapplet");
  }
}

KAntsApplet::KAntsApplet(const QString& configFile,
			 Type type, int actions,
			 QWidget *parent, const char *name)
  : KPanelApplet(configFile, type, actions, parent, name)
{
  //
  // Setup
  //
  aboutData = 0;
  usock = -1;
  curJob = jobs.end();
  preferredJobType = "None";
  //
  // Fetch the configuration
  //
  config = new KSimpleConfig(configFile);

  if (config->hasKey("preferredJobType")) {
    preferredJobType = config->readEntry("preferredJobType");
  }

  //
  // Setup the applet's style.
  //
  setBackgroundColor(black);
  setFrameStyle(StyledPanel | Sunken);

  connectSocket();
}

int KAntsApplet::widthForHeight(int h) const
{
  int size = 0;
  int offset = BAR_MARGIN;
  int cellSize;
  if (curJob != jobs.end()) {  
    horzBarValues (h, offset, cellSize);
    
    size = offset + (cellSize + BAR_MARGIN) * curJob->second.jobStats.size();
  } 

  if (size > h)
    return size;
  else
    return h;
}

int KAntsApplet::heightForWidth(int w) const
{
  int size = 0;

  if (curJob != jobs.end()) {
    size = 2 * (TEXT_MARGIN + frameWidth()) 
      + vertBarValues(w) * curJob->second.jobStats.size();

    QPainter p (this);
    QFont font("fixed");
    font.setPixelSize(TEXT_SIZE);
    p.setFont(font);
    
    QRect rect = p.boundingRect (TEXT_MARGIN / 2, TEXT_MARGIN / 2, 
				 TEXT_MARGIN / 2, TEXT_MARGIN / 2,
				 AlignLeft, curJob->first);

    size += rect.height();
  }

  return std::max (w, size);
}

void KAntsApplet::dataReceived()
{
  char buf[LOG_LEN];
  unsigned peerId;
  char jobtype[TYPE_LEN];
  AntsMember item;
  int rc;

  //
  // Receive data from ants
  //
  rc = recv(usock, buf, LOG_LEN, 0);
  if (rc <= 0)
    closeSocket();

  rc = sscanf(buf, "%u %15s %u %u %u %u",
	      &peerId, jobtype, &item.max,
	      &item.preferred, &item.running, 
	      &item.reserved);

  if (rc != 6) 
    return;

  //
  // See if we can locate the jobtype.
  //
  AntsJobs::iterator job = jobs.find(jobtype);

  if(job == jobs.end()) {
    //
    // If the jobtype didn't exist, we just create it.
    //
    AntsJob newjob;
    newjob.max = 0;
    job = jobs.insert (std::make_pair(QString(jobtype), newjob)).first;
  } 

  AntsJobStats::iterator stats = job->second.jobStats.find(peerId);
  
  if (stats == job->second.jobStats.end()) {
    //
    // This is a new peer. Insert it.
    //
    job->second.jobStats.insert(std::make_pair(peerId, item));
    job->second.max = std::max (item.max, job->second.max);
    emit updateLayout();
  } else { 
    stats->second = item;
  }

  if (QString(jobtype) == preferredJobType)
    curJob = job;

  if (curJob == jobs.end())
    curJob = jobs.begin();

  if (curJob == job)
    update();
}

void KAntsApplet::about()
{
  if (!aboutData) {
    aboutData = new KAboutData("kantsapplet", "KAntsApplet",
			       ANTS_VERSION, "KDE panel-applet for "
			       "the Ants monitoring system.",
			       KAboutData::License_GPL_V2, "(C) 2002, "
			       "Klaus S. Madsen");
    aboutData->addAuthor("Klaus S. Madsen", 0, "ksm@hjernemadsen.org");
  }
  KAboutApplication dialog(aboutData);
  dialog.show();
}

void KAntsApplet::drawContents(QPainter* p)
{
  QRect r = contentsRect();

  //
  // Select the correct font.
  //
  QFont font("fixed");
  font.setPixelSize(TEXT_SIZE);

  //
  // Write out the current jobtype
  //
  p->setFont(font);
  p->setPen(white);

  if (usock < 0) {
    p->drawText(TEXT_MARGIN / 2, TEXT_SIZE + TEXT_MARGIN / 2, "(error)");
  } else {
    if (curJob != jobs.end())
      p->drawText(TEXT_MARGIN / 2, TEXT_SIZE + TEXT_MARGIN / 2, curJob->first);
    else
      p->drawText(TEXT_MARGIN / 2, TEXT_SIZE + TEXT_MARGIN / 2, "(no data)");
  }

  if (curJob != jobs.end()) {
    if (orientation() == Horizontal) {
      int xpos = BAR_MARGIN;
      int cellSize;
      
      horzBarValues (r.height(), xpos, cellSize);
      //
      //
      //
      for (AntsJobStats::const_iterator i = curJob->second.jobStats.begin() 
	     ; i != curJob->second.jobStats.end() ; ++i) {
	int ypos = r.height() - BAR_MARGIN / 2 - cellSize;
	p->setPen(NoPen);

	//
	// Draw the cells for the processes that are running, but not 
	// overcommitted.
	//
	p->setBrush(green);
	drawHorzCells(p, std::min(i->second.running, i->second.preferred), cellSize,
		      xpos, ypos);

	//
	// Draw the overcomitted cells.
	//
	p->setBrush(red);
	drawHorzCells(p, i->second.running - i->second.preferred, cellSize,
		      xpos, ypos);

	//
	// Draw the reserved cells.
	//
	p->setBrush(blue);
	drawHorzCells(p, i->second.reserved, cellSize, xpos, ypos);
	
	//
	// Draw the rest.
	//
	p->setBrush(gray);
	drawHorzCells(p, i->second.max - i->second.reserved 
		      - i->second.running,
		      cellSize, xpos, ypos);

	xpos += BAR_MARGIN + cellSize;
      }  
    } else {
      int ypos = TEXT_SIZE + TEXT_MARGIN;
      int cellSize = vertBarValues (r.width());
      
      //
      //
      //
      for (AntsJobStats::const_iterator i = curJob->second.jobStats.begin() 
	     ; i != curJob->second.jobStats.end() ; ++i) {
	int xpos = BAR_MARGIN;
	p->setPen(NoPen);

	//
	// Draw the cells for the processes that are running, but not 
	// overcommitted.
	//
	p->setBrush(green);
	drawVertCells(p, std::min(i->second.running, i->second.preferred), cellSize,
		      ypos, xpos);

	//
	// Draw the overcomitted cells.
	//
	p->setBrush(red);
	drawVertCells(p, i->second.running - i->second.preferred, cellSize,
		      ypos, xpos);

	//
	// Draw the reserved cells.
	//
	p->setBrush(blue);
	drawVertCells(p, i->second.reserved, cellSize, ypos, xpos);
	
	//
	// Draw the rest.
	//
	p->setBrush(gray);
	drawVertCells(p, i->second.max - i->second.reserved 
		      - i->second.running, 
		      cellSize, ypos, xpos);

	ypos += BAR_MARGIN + cellSize;
      }
    }
  }
}
void KAntsApplet::mousePressEvent(QMouseEvent *e)
{
  if (e->button() == QMouseEvent::LeftButton) {
    if (usock < 0) {
      connectSocket();
      return;
    }

    curJob++;
    if (curJob == jobs.end())
      curJob = jobs.begin();

    update();
    emit updateLayout();
  } else if (e->button() == QMouseEvent::RightButton) {
    if (jobs.size()) {
      QPopupMenu menu (this);
      menu.setCheckable(TRUE);
      QMap<int, QString> menuItems;
      
      int menuId = 1;
      for (AntsJobs::iterator i = jobs.begin() ; i != jobs.end(); i++) {
	int id = menu.insertItem(i->first, menuId);
	if (i == curJob)
	  menu.setItemChecked(id, TRUE);
	
	menuItems[menuId++] = i->first;
      }
      menu.adjustSize();
      
      int res = menu.exec(mapToGlobal(e->pos()));

      if (menuItems[res] != "") {
	AntsJobs::iterator i = jobs.find (menuItems[res]);
	if (i != jobs.end()) {
	  curJob = i;
	}
	update();
	emit updateLayout();
      }
    }
  }
  if (curJob != jobs.end()) {
    preferredJobType = curJob->first;
  }
    
}

void KAntsApplet::horzBarValues(int h, int &offset, int &cellSize) const
{
  //
  // If there isn't room for the bars under the text, 
  // the bars should start only when they are free of the text.
  //
  int max = curJob->second.max;

  if (max) {
    cellSize = (h - TEXT_SIZE - TEXT_MARGIN) / max;

    if (cellSize <= LOW_CELLSIZE) {
      //
      // The bars reach the text. Damn. 
      // Use the smallest cellSize
      //
      cellSize = LOW_CELLSIZE;

      QPainter p (this);
      QFont font("fixed");
      font.setPixelSize(TEXT_SIZE);
      p.setFont(font);

      QRect rect = p.boundingRect (TEXT_MARGIN / 2, TEXT_MARGIN / 2, 
				   h - TEXT_MARGIN / 2, h - TEXT_MARGIN / 2,
				   AlignLeft, curJob->first);


      offset += rect.width() + TEXT_MARGIN / 2;
    } else {
      //
      // The cells fit.
      //
      cellSize--;
    }
  }
}

void KAntsApplet::drawHorzCells(QPainter* p, int cells, int size, int xpos, int &ypos)
{
  for ( ; cells > 0 ; cells--) {
    p->drawRect(xpos, ypos, size, size);
    ypos -= size + 1;
  }
}

int KAntsApplet::vertBarValues(int w) const
{
  int max = curJob->second.max;
  int cellSize = 0;

  if (max) {
    cellSize = (w - 2 * (BAR_MARGIN + frameWidth())) / max;

    if (cellSize <= LOW_CELLSIZE)
      cellSize = LOW_CELLSIZE;
  }
  return cellSize;
}

void KAntsApplet::drawVertCells(QPainter* p, int cells, int size, int ypos, int &xpos)
{
  for ( ; cells > 0 ; cells--) {
    p->drawRect(xpos, ypos, size, size);
    xpos += size + 1;
  }
}

void KAntsApplet::connectSocket()
{
  if (usock < 0) {
    usock = ants_local_connect();
    if (usock < 0) {
      jobs.clear();
      curJob = jobs.end();
      usock = -1;
      return;
    }
    
    int rc = ants_statclient(usock);
    if (rc != 0) {
      close (usock);
      usock = -1;
      jobs.clear();
      curJob = jobs.end();
    }
    
    notify = new QSocketNotifier(usock, QSocketNotifier::Read, this);
    QObject::connect(notify, SIGNAL(activated(int)),
		     this, SLOT(dataReceived()));
  }
}

void KAntsApplet::closeSocket()
{
  delete notify;
  close (usock);
  usock = -1;

  jobs.clear();
  curJob = jobs.end();
 
  emit updateLayout();
}
